
#include <stdio.h>
#include <signal.h>

#include <va/va_x11.h>

#include <libavcodec/avcodec.h>
#include <libavcodec/vaapi.h>
#include <libavformat/avformat.h>

#include <termios.h>
#include <unistd.h>
#include <sys/time.h>

#include "va.h"

#define OFFSDELTA 20

int quit_request = 0;


static void exit_sighandler(int x)
{
	switch(x)
	{
	case SIGINT:
	case SIGQUIT:
	case SIGTERM:
	case SIGKILL:
		quit_request = 1;
		return;
	default:;
	}
	
	exit(1);
}


static const char *VAProfile2string(VAProfile profile)
{
	switch(profile)
	{
#define PROFILE(profile) \
	case VAProfile##profile: return "VAProfile" #profile
	PROFILE(MPEG2Simple);
	PROFILE(MPEG2Main);
	PROFILE(MPEG4Simple);
	PROFILE(MPEG4AdvancedSimple);
	PROFILE(MPEG4Main);
	PROFILE(H264Baseline);
	PROFILE(H264Main);
	PROFILE(H264High);
	PROFILE(VC1Simple);
	PROFILE(VC1Main);
	PROFILE(VC1Advanced);
#undef PROFILE
	default: break;
	}
	return "<unknown>";
}

static const char *VAEntrypoint2string(VAEntrypoint entrypoint)
{
	switch(entrypoint)
	{
#define ENTRYPOINT(entrypoint) \
	case VAEntrypoint##entrypoint: return "VAEntrypoint" #entrypoint
	ENTRYPOINT(VLD);
	ENTRYPOINT(IZZ);
	ENTRYPOINT(IDCT);
	ENTRYPOINT(MoComp);
	ENTRYPOINT(Deblocking);
#undef ENTRYPOINT
	default: break;
	}
	return "<unknown>";
}


#define IMGFMT_NV12 0x3231564E
#define IMGFMT_YV12 0x32315659
#define IMGFMT_I420 0x30323449
#define IMGFMT_IYUV 0x56555949

#define IMGFMT_VAAPI               0x56410000 /* 'VA'00 */
#define IMGFMT_VAAPI_CODEC_MPEG2   (0x10)
#define IMGFMT_VAAPI_CODEC_MPEG4   (0x20)
#define IMGFMT_VAAPI_CODEC_H264    (0x30)
#define IMGFMT_VAAPI_CODEC_VC1     (0x40)
#define IMGFMT_VAAPI_MPEG2         (IMGFMT_VAAPI | IMGFMT_VAAPI_CODEC_MPEG2)
#define IMGFMT_VAAPI_MPEG2_IDCT    (IMGFMT_VAAPI | IMGFMT_VAAPI_CODEC_MPEG2 | 1)
#define IMGFMT_VAAPI_MPEG2_MOCO    (IMGFMT_VAAPI | IMGFMT_VAAPI_CODEC_MPEG2 | 2)
#define IMGFMT_VAAPI_MPEG4         (IMGFMT_VAAPI | IMGFMT_VAAPI_CODEC_MPEG4)
#define IMGFMT_VAAPI_H263          (IMGFMT_VAAPI | IMGFMT_VAAPI_CODEC_MPEG4 | 1)
#define IMGFMT_VAAPI_H264          (IMGFMT_VAAPI | IMGFMT_VAAPI_CODEC_H264)
#define IMGFMT_VAAPI_VC1           (IMGFMT_VAAPI | IMGFMT_VAAPI_CODEC_VC1)
#define IMGFMT_VAAPI_WMV3          (IMGFMT_VAAPI | IMGFMT_VAAPI_CODEC_VC1 | 1)

static const struct
{
    int fmt;
    enum PixelFormat pix_fmt;
    enum CodecID codec_id;

} conversion_map[] = 
{
	{IMGFMT_NV12,  PIX_FMT_NV12},
	{IMGFMT_YV12,  PIX_FMT_YUV420P},
	{IMGFMT_I420,  PIX_FMT_YUV420P},
	{IMGFMT_IYUV,  PIX_FMT_YUV420P},


	{IMGFMT_VAAPI_MPEG2,     PIX_FMT_VAAPI_VLD,  CODEC_ID_MPEG2VIDEO},
	{IMGFMT_VAAPI_MPEG2_IDCT,PIX_FMT_VAAPI_IDCT, CODEC_ID_MPEG2VIDEO},
	{IMGFMT_VAAPI_MPEG2_MOCO,PIX_FMT_VAAPI_MOCO, CODEC_ID_MPEG2VIDEO},
	{IMGFMT_VAAPI_MPEG4,     PIX_FMT_VAAPI_VLD,  CODEC_ID_MPEG4},
	{IMGFMT_VAAPI_H263,      PIX_FMT_VAAPI_VLD,  CODEC_ID_H263},
	{IMGFMT_VAAPI_H264,      PIX_FMT_VAAPI_VLD,  CODEC_ID_H264},
	{IMGFMT_VAAPI_WMV3,      PIX_FMT_VAAPI_VLD,  CODEC_ID_WMV3},
	{IMGFMT_VAAPI_VC1,       PIX_FMT_VAAPI_VLD,  CODEC_ID_VC1},

	{0, PIX_FMT_NONE}
};

enum PixelFormat imgfmt2pixfmt(int fmt)
{
	int i;
	enum PixelFormat pix_fmt;
	for(i = 0; conversion_map[i].fmt; i++)
	{
		if(conversion_map[i].fmt == fmt)
			break;
	}
	pix_fmt = conversion_map[i].pix_fmt;
	if(pix_fmt == PIX_FMT_NONE)
		printf("ERROR Unsupported format %d\n", fmt);
	return pix_fmt;
}

int pixfmt2imgfmt(enum PixelFormat pix_fmt, int codec_id)
{
	int i;
	int fmt;
	for(i = 0; conversion_map[i].pix_fmt != PIX_FMT_NONE; i++)
	{
		//printf("pixfmt2imgfmt: #%d  pix fmt 0x%08X, 0x%08X  codec %d, %d\n", i, pix_fmt, conversion_map[i].pix_fmt, codec_id, conversion_map[i].codec_id); 
		if((conversion_map[i].pix_fmt == pix_fmt) && ((conversion_map[i].codec_id == 0) || (conversion_map[i].codec_id == codec_id)))
			break;
	}
	fmt = conversion_map[i].fmt;
	if(!fmt)
		printf("ERROR Unsupported PixelFormat %i\n", pix_fmt);
	return fmt;
}

typedef struct
{
	VASurfaceID id;
	VAImage* img;
	char used;
	char bound;

} surface_t;

typedef struct
{
	void* this;
	uint8_t* indata;
	uint8_t* data;
	int bsize;
	int inwpos;
	int inrpos;
	int pos;

}dbuf_t;

typedef struct
{
	char fatal_error;
	VADisplay* vadpy;
	dbuf_t	buf;
	FILE* file;
	AVCodecContext* avctxt;
	AVFormatContext *ic;
	AVInputFormat *format;
	ByteIOContext* ioctxt;
	char seekable;
	struct vaapi_context *hwaccel;
	AVFrame* avframe;
	int wnd_w;
	int wnd_h;
	int v_width;
	int v_height;
	double v_ratio;
	int csp;
	int imgfmt;
	enum PixelFormat pixfmt_default;
	VAImageFormat va_imgfmt;
	char hw;
	int surface_count;
	char fmt_changed;
	surface_t* surfaces;
	VASurfaceID* surface_ids;
	int surface_idx;
	surface_t* sw_surface;

} ctxt_t;

static int init_vo(ctxt_t *ctxt, int fmt, VAEntrypoint ep, VAProfile prfl);

const char *VAImageFormat2str(VAImageFormat *imgfmt)
{
    static char str[5];
    str[0] = imgfmt->fourcc;
    str[1] = imgfmt->fourcc >> 8;
    str[2] = imgfmt->fourcc >> 16;
    str[3] = imgfmt->fourcc >> 24;
    str[4] = '\0';
    return str;
}

#define fast_memcpy(a,b,c) memcpy(a,b,c)

static inline void * memcpy_pic(void * dst, const void * src, int bytesPerLine, int height, int dstStride, int srcStride)
{
	int i;
	void *retval = dst;

	if(dstStride == srcStride)
	{
		if(srcStride < 0)
		{
			src = (uint8_t*)src + (height - 1) * srcStride;
			dst = (uint8_t*)dst + (height - 1) * dstStride;
			srcStride = -srcStride;
		}

		fast_memcpy(dst, src, srcStride * height);
	}
	else
	{
		for(i = 0; i < height; i++)
		{
			fast_memcpy(dst, src, bytesPerLine);
			src = (uint8_t*)src + srcStride;
			dst = (uint8_t*)dst + dstStride;
		}
	}

	return retval;
}


#define IMGTYPE_TEMP			1
#define IMGTYPE_STATIC			2

#define IMGFLAG_READABLE		1
#define IMGFLAG_PRESERVE		2
#define IMGFLAG_DRAW_CALLBACK	4

static surface_t* get_surface(ctxt_t* ctxt)
{
	int idx = ctxt->surface_idx;
	for(;;)
	{
		if(ctxt->surface_idx == ctxt->surface_count)
			ctxt->surface_idx = 0;

		printf("get_srfc #%d\n",ctxt->surface_idx);
		if(!ctxt->surfaces[ctxt->surface_idx].used)
		{
			ctxt->surfaces[ctxt->surface_idx].used = 1;
			break;
		}

		ctxt->surface_idx++;

		if(ctxt->surface_idx == idx)
		{
			printf("ERROR get surface\n");
			return NULL;
		}
	}

	//printf("get_surface #%d\n", ctxt->surface_idx);
	return &ctxt->surfaces[ctxt->surface_idx];
}

static int av_get_buffer(AVCodecContext *avctxt, AVFrame *pic)
{
	ctxt_t* ctxt = (ctxt_t*)avctxt->opaque;

	printf("get_buffer buffer_hints: %d\n", pic->buffer_hints);
	
	surface_t* srfc = get_surface(ctxt);
	if(!srfc)
		return -1;

	pic->type = FF_BUFFER_TYPE_USER;
	pic->age = 1;
	pic->data[0] = (void*)&srfc->id;
	pic->data[1] = NULL;
	pic->data[2] = NULL;
	pic->data[3] = (void*)(size_t)srfc->id;
	pic->linesize[0] = 0;
	pic->linesize[1] = 0;
	pic->linesize[2] = 0;
	pic->linesize[3] = 0;

	printf("av_get_buffer done pic %p d[0] %p (0x%08X) d[3] 0x%08X\n", pic, pic->data[0], (int)*(VASurfaceID*)pic->data[0], (int)(size_t)pic->data[3]);

	return 0;
}

static void av_release_buffer(AVCodecContext *avctx, AVFrame *pic)
{
	ctxt_t* ctxt = (ctxt_t*)avctx->opaque;
	VASurfaceID sid = (VASurfaceID)(size_t)pic->data[3];
	int i = 0;

	printf("release_buffer 0x%08X\n", sid);
	
	for(i = 0; i < ctxt->surface_count; i++)
	{
		//printf("srfc #%d %p 0x%08X\n",i,srfcs->srfc_used[i], srfcs->srfc_used[i] ? *srfcs->srfc_used[i] : -1);
		if(ctxt->surfaces[i].id == sid)
		{
			ctxt->surfaces[i].used = 0;
			break;
		}
	}

	pic->data[0] = NULL;
	pic->data[1] = NULL;
	pic->data[2] = NULL;
	pic->data[3] = NULL;
}


static int draw_slice(ctxt_t *ctxt, uint8_t * image[], int stride[], int w, int h, int x, int y)
{
	surface_t* const surface = ctxt->sw_surface;
	VAImage* const va_image = surface ? surface->img : NULL;
	VAStatus status;
	uint8_t *image_data = NULL;
	uint8_t *dst[3] = { 0, };
	unsigned int dst_stride[3] = { 0, 0, 0, };

	printf("draw_slice(): (%d) location (%d,%d), size %dx%d  srfc 0x%08X\n", !ctxt->hw, x, y, w, h, surface ? surface->id : -1);

	if(ctxt->hw)
		return 1;

	if(!surface)
	{
		printf("ERROR draw_slice: no surface!\n");
		return 0;
	}

	if((status = vaMapBuffer(ctxt->vadpy, va_image->buf, (void *)&image_data)) != VA_STATUS_SUCCESS)
	{
		printf("ERROR draw_slice vaMapBuffer: %d\n", status);
		return 0;
	}

	dst_stride[0] = va_image->pitches[0];
	dst[0] = image_data + va_image->offsets[0] + y * dst_stride[0] + x;

	memcpy_pic(dst[0], image[0], w, h, dst_stride[0], stride[0]);

	x /= 2;
	y /= 2;
	w /= 2;
	h /= 2;

	if(ctxt->imgfmt == IMGFMT_YV12)
	{
		dst_stride[1] = va_image->pitches[2];
		dst[1] = image_data + va_image->offsets[2] + y * dst_stride[1] + x;
		dst_stride[2] = va_image->pitches[1];
		dst[2] = image_data + va_image->offsets[1] + y * dst_stride[2] + x;
	}
	else
	{
		if(image[1])
		{
			dst_stride[1] = va_image->pitches[1];
			dst[1] = image_data + va_image->offsets[1] + y * dst_stride[1] + x;
		}
		if(image[2])
		{
			dst_stride[2] = va_image->pitches[2];
			dst[2] = image_data + va_image->offsets[2] + y * dst_stride[2] + x;
		}
	}

	if(image[1])
		memcpy_pic(dst[1], image[1], w, h, dst_stride[1], stride[1]);

	if(image[2])
		memcpy_pic(dst[2], image[2], w, h, dst_stride[2], stride[2]);

	if((status = vaUnmapBuffer(ctxt->vadpy, surface->img->buf)) != VA_STATUS_SUCCESS)
	{
		printf("ERROR draw_slice  vaUnmapBuffer: %d", status);
		return 0;
	}

	return 1;
}


static void set_format_params(struct AVCodecContext *avctxt, enum PixelFormat fmt)
{
	int imgfmt;

	if(fmt == PIX_FMT_NONE)
		return;

	imgfmt = pixfmt2imgfmt(fmt, avctxt->codec_id);
	printf("set_format_params: IMGFMT_VAAPI %d\n", (imgfmt & 0xFFFF0000) == IMGFMT_VAAPI);

	avctxt->thread_count    = 1;
	avctxt->get_buffer = avcodec_default_get_buffer;
	avctxt->reget_buffer = avcodec_default_reget_buffer;
	avctxt->release_buffer = avcodec_default_release_buffer;
	avctxt->draw_horiz_band = NULL;
	avctxt->slice_flags = SLICE_FLAG_CODED_ORDER | SLICE_FLAG_ALLOW_FIELD;

	if((imgfmt & 0xFFFF0000) == IMGFMT_VAAPI)
	{
		avctxt->get_buffer      = av_get_buffer;
		avctxt->release_buffer  = av_release_buffer;
		avctxt->reget_buffer    = av_get_buffer;
	}

}

static VAProfile find_profile(VAProfile* lp, int np, int p)
{
	int i;
	for(i = 0; i < np; i++)
	{
		if(lp[i] == p)
			return lp[i];
	}
	
	return -1;
}

static void create_surfaces(ctxt_t *ctxt)
{
	ctxt->surface_count = ctxt->hw ? (ctxt->avctxt->refs + 19) : 1;//ctxt->avctxt->refs + (ctxt->hw ? 2 : 2);
	printf("init_vo: create %d surfaces  %d x %d...\n", ctxt->surface_count, ctxt->avctxt->width, ctxt->avctxt->height);
	ctxt->surfaces = malloc(sizeof(surface_t) * (ctxt->surface_count));
	ctxt->surface_idx = 0;

	VAStatus status = VA_STATUS_SUCCESS;
	int ret = 0;
	int i;
	for(i = 0; i < ctxt->surface_count; i++)
	{
		ctxt->surfaces[i].id = VA_INVALID_ID;
		ctxt->surfaces[i].img = NULL;
		ctxt->surfaces[i].used = 0;
		ctxt->surfaces[i].bound = 0;
	}

	if(ctxt->surface_ids)
		free(ctxt->surface_ids);
	ctxt->surface_ids = malloc(sizeof(VASurfaceID) * ctxt->surface_count);

	if((status == VA_STATUS_SUCCESS) && ((status = vaCreateSurfaces(ctxt->vadpy, ctxt->avctxt->width, ctxt->avctxt->height, VA_RT_FORMAT_YUV420, ctxt->surface_count, ctxt->surface_ids)) != VA_STATUS_SUCCESS))
		printf("ERROR init_vo: create surfaces: %d\n", status);
	else
	{
		for(i = 0; i < ctxt->surface_count; i++)
		{
			printf("create_surfaces #%d 0x%08X\n", i, ctxt->surface_ids[i]);
			ctxt->surfaces[i].id = ctxt->surface_ids[i];

			if(!ctxt->hw)
			{
				ctxt->surfaces[i].img = malloc(sizeof(VAImage));
				if((status = vaDeriveImage(ctxt->vadpy, ctxt->surfaces[i].id, ctxt->surfaces[i].img)) == VA_STATUS_SUCCESS)
				{
					if(ctxt->surfaces[i].img->format.fourcc != ctxt->va_imgfmt.fourcc)
					{
						printf("ERROR init_vo vaDeriveImage: fourcc (0x%80X  0x%08X)\n", ctxt->surfaces[i].img->format.fourcc, ctxt->va_imgfmt.fourcc);
						ret = -1;
					}
					else if((ctxt->surfaces[i].img->width != ctxt->avctxt->width) || (ctxt->surfaces[i].img->height != ctxt->avctxt->height))
					{
						printf("ERROR init_vo vaDeriveImage: (%d x %d, %d x %d)\n", ctxt->surfaces[i].img->width, ctxt->surfaces[i].img->height, ctxt->avctxt->width, ctxt->avctxt->height);
						ret = -1;
					}

					if(ret < 0)
						vaDestroyImage(ctxt->vadpy, ctxt->surfaces[i].img->image_id);
					else
						ctxt->surfaces[i].bound = 1;
				}
				else
					ret = -1;

				if(ret < 0)
				{
					printf("WARN  init_vo vaDeriveImage: %d\n",status);
					ret = 0;
					if((status = vaCreateImage(ctxt->vadpy, &ctxt->va_imgfmt, ctxt->avctxt->width, ctxt->avctxt->height, ctxt->surfaces[i].img)) != VA_STATUS_SUCCESS)
					{
						printf("ERROR init_vo vaCreateImage: %d\n", status);
						ret = -1;
					}
					else
						printf("srfc->img 0x%08X\n", ctxt->surfaces[i].img->image_id);
				}

				if(ret < 0)
					break;
			}
		}
	}

	if(!ctxt->hw)
		ctxt->sw_surface = get_surface(ctxt);
}

static void destroy_surfaces(ctxt_t *ctxt)
{
	if(ctxt->surfaces)
	{
		int i;
		for(i =0 ; i < ctxt->surface_count; i++)
		{
			if(ctxt->surfaces[i].img)
			{
				printf("destroying image 0x%08X\n", ctxt->surfaces[i].img->image_id);
				vaDestroyImage(ctxt->vadpy, ctxt->surfaces[i].img->image_id);
				free(ctxt->surfaces[i].img);
			}
			printf("destroying surface 0x%08X\n", ctxt->surfaces[i].id);
			vaDestroySurfaces(ctxt->vadpy, &ctxt->surfaces[i].id, 1);
		}

		free(ctxt->surfaces);
		ctxt->surfaces = NULL;
	}

	if(ctxt->surface_ids)
		free(ctxt->surface_ids);
	ctxt->surface_ids = NULL;
	
	ctxt->sw_surface = NULL;
}

static void reset_buf(ctxt_t* this)
{
	this->buf.inwpos = 0;
	this->buf.inrpos = 0;
	this->buf.bsize = 0;
	this->buf.pos = 0;
}

static void close_avctxt(ctxt_t *ctxt)
{
	printf("close_avctxt\n");

	if(ctxt->avctxt)
	{
		if(ctxt->avctxt->codec)
		{
			printf("close avcodec %p\n", ctxt->avctxt);
			avcodec_close(ctxt->avctxt);
		}
		ctxt->avctxt->hwaccel_context = NULL;
		ctxt->avctxt = NULL;
	}

	if(ctxt->ic)
		av_close_input_stream(ctxt->ic);
	ctxt->ic = NULL;

	if(ctxt->ioctxt)
		av_free(ctxt->ioctxt);
	ctxt->ioctxt = NULL;
	ctxt->format = NULL;

	if(ctxt->file)
		fclose(ctxt->file);
	ctxt->file = NULL;

	if(ctxt->avframe)
		av_free(ctxt->avframe);
	ctxt->avframe = NULL;

	reset_buf(ctxt);
}

static void free_vo(ctxt_t *ctxt)
{
	printf("free_vo\n");

	if(ctxt->hwaccel)
	{
		if(ctxt->hwaccel->context_id)
			vaDestroyContext(ctxt->vadpy, ctxt->hwaccel->context_id);
		ctxt->hwaccel->context_id = 0;

		if(ctxt->hwaccel->config_id)
			vaDestroyConfig(ctxt->vadpy, ctxt->hwaccel->config_id);
		ctxt->hwaccel->config_id = 0;
	}

	ctxt->hw = 0;
	ctxt->v_ratio = 0.0;
	ctxt->v_width = 0;
	ctxt->v_height = 0;

}

static char get_entrypoint(ctxt_t *ctxt, VAEntrypoint* ep, VAProfile* prfl)
{
	VAStatus status;
	AVCodecContext *avctxt = ctxt->avctxt;
	*prfl = -1;

	if(!avctxt)
		return 0;

	int np;
	VAProfile* p = malloc(vaMaxNumProfiles(ctxt->vadpy) * sizeof(VAProfile));

	if((status = vaQueryConfigProfiles(ctxt->vadpy, p, &np)) == VA_STATUS_SUCCESS)
	{
		switch(avctxt->codec_id)
		{
		case CODEC_ID_MPEG2VIDEO: *prfl = find_profile(p, np, VAProfileMPEG2Main); break;
		//case CODEC_ID_MPEG2VIDEO: psel = find_profile(p, np, VAProfileMPEG2Simple); break;
		case CODEC_ID_MPEG4:
		case CODEC_ID_H263: *prfl = find_profile(p, np, VAProfileMPEG4AdvancedSimple); break;
		case CODEC_ID_H264:
			{
				if(avctxt->profile == FF_PROFILE_H264_BASELINE)
				{
					*prfl = find_profile(p, np, VAProfileH264Baseline);
					if(*prfl == -1)
						*prfl = find_profile(p, np, VAProfileH264Main);
					if(*prfl == -1)
						*prfl = find_profile(p, np, VAProfileH264High);
				}
				else
				{
					if(avctxt->profile == FF_PROFILE_H264_MAIN)
						*prfl = find_profile(p, np, VAProfileH264Main);
					if(*prfl == -1)
						*prfl = find_profile(p, np, VAProfileH264High);
				}
				break;
			}
		case CODEC_ID_WMV3: *prfl = find_profile(p, np, VAProfileVC1Main); break;
		case CODEC_ID_VC1: *prfl = find_profile(p, np, VAProfileVC1Advanced); break;
		default:;
		}

		if(*prfl == -1)
			printf("ERROR find profiles (p %d codec %d)!\n", avctxt->profile, avctxt->codec_id);
	}
	else
		printf("ERROR vaQueryConfigProfiles: %d (dpy %p)\n", status, ctxt->vadpy);

	if(p)
		free(p);

	if(*prfl == -1)
		status = VA_STATUS_ERROR_UNKNOWN;

	*ep = 0;
	if(status == VA_STATUS_SUCCESS)
	{
		VAEntrypoint* eps = malloc(vaMaxNumEntrypoints(ctxt->vadpy) * sizeof(VAEntrypoint));
		int neps;

		if((status = vaQueryConfigEntrypoints(ctxt->vadpy, *prfl, eps, &neps)) != VA_STATUS_SUCCESS)
			printf("ERROR vaQueryConfigEntrypoints: %d\n", status);

		if(status == VA_STATUS_SUCCESS)
			*ep = eps[0];

		free(eps);
	}
	//*ep=0;
	printf("Profile: %d (%s) Entrypoint %d (%s)\n", *prfl, VAProfile2string(*prfl), *ep, VAEntrypoint2string(*ep));

	if(*ep != VAEntrypointVLD)
	{
		*ep = 0;
		printf("ERROR only 'VAEntrypointVLD' supported\n");
	}

	return (*ep > 0) ? 1 : 0;
}

static int init_vo(ctxt_t *ctxt, int fmt, VAEntrypoint ep, VAProfile prfl)
{
	AVCodecContext *avctxt = ctxt->avctxt;
	double ratio;
	VAStatus status = VA_STATUS_SUCCESS;
	int imgfmt;
	VAImageFormat* ifmts = NULL;
	VAImageFormat* va_imgfmt = NULL;
	int n_ifmts;
	int i;
	int ret = 0;

	if(!avctxt || !avctxt->codec)
	{
		printf("ERROR no avctxt (%p) / codec (%p)\n", avctxt, avctxt ? avctxt->codec : NULL);
		return -1;
	}

	ratio = avctxt->width / (double)avctxt->height;
	ratio *= (avctxt->sample_aspect_ratio.num ? avctxt->sample_aspect_ratio.num : 1) / (double)avctxt->sample_aspect_ratio.den;

	//printf("---------> avctxt->sample_aspect_ratio %d %d -> %f\n",avctxt->sample_aspect_ratio.num,avctxt->sample_aspect_ratio.den,ratio);	
	if(((int)(ctxt->v_ratio * 10) != (int)(ratio * 10)) || (ctxt->v_width != avctxt->width) || (ctxt->v_height != avctxt->height))
	{
		printf("init_vo w %d -> %d h %d -> %d r %f -> %f  \n", ctxt->v_width, avctxt->width, ctxt->v_height, avctxt->height, ctxt->v_ratio, ratio);
		free_vo(ctxt);

		destroy_surfaces(ctxt);

		ifmts = calloc(vaMaxNumImageFormats(ctxt->vadpy), sizeof(VAImageFormat));
		if(ifmts)
			vaQueryImageFormats(ctxt->vadpy, ifmts, &n_ifmts);

		imgfmt = pixfmt2imgfmt(fmt, avctxt->codec_id);
		switch(imgfmt)
		{
		case IMGFMT_VAAPI_MPEG2:
		case IMGFMT_VAAPI_MPEG4:
		case IMGFMT_VAAPI_H263:
		case IMGFMT_VAAPI_H264:
		case IMGFMT_VAAPI_WMV3:
		case IMGFMT_VAAPI_VC1:
			break;
		case IMGFMT_NV12:
		case IMGFMT_YV12:
		case IMGFMT_I420:
		case IMGFMT_IYUV:
			for(i = 0; i < n_ifmts; i++)
			{
				if(ifmts[i].fourcc == imgfmt)
				{
					va_imgfmt = &ifmts[i];
					break;
				}
			}
			if(va_imgfmt)
				break;
		default:
			printf("ERROR init_vo: format 0x%08X not supported\n", fmt);
			ret = -1;
		}
	
		if(va_imgfmt)
			ctxt->va_imgfmt = *va_imgfmt;
		else
		{
			int dimgfmt = pixfmt2imgfmt(ctxt->pixfmt_default, avctxt->codec_id);
			for(i = 0; i < n_ifmts; i++)
			{
				if(ifmts[i].fourcc == dimgfmt)
				{
					ctxt->va_imgfmt = ifmts[i];
					break;
				}
			}
		
		}

		if(ret == 0)
		{
			ctxt->hw = ((imgfmt & 0xFFFF0000) == IMGFMT_VAAPI);
			
			printf("init_vo: pixfmt 0x%08X, imgfmt 0x%08X, va_imgfmt '%s', hw %d\n", fmt, imgfmt, VAImageFormat2str(&ctxt->va_imgfmt), ctxt->hw);

			create_surfaces(ctxt);
		}

		if((ret == 0) && ctxt->hw)
		{
			VAConfigAttrib attrib;
			attrib.type = VAConfigAttribRTFormat;

			
			if((status == VA_STATUS_SUCCESS) && (status = vaGetConfigAttributes(ctxt->vadpy, prfl, ep, &attrib, 1)) != VA_STATUS_SUCCESS)
				printf("ERROR init_vo: get config!\n");

			if((status == VA_STATUS_SUCCESS) && (attrib.value & VA_RT_FORMAT_YUV420) == 0)
			{
				status = VA_STATUS_ERROR_UNKNOWN;
				printf("ERROR init_vo: invalid yuv format %x", attrib.value);
			}

			if((status == VA_STATUS_SUCCESS) && ((status = vaCreateConfig(ctxt->vadpy, prfl, ep, &attrib, 1, &ctxt->hwaccel->config_id)) != VA_STATUS_SUCCESS))
				printf("ERROR init_vo: create config!\n");

			if((status == VA_STATUS_SUCCESS) && (status = vaCreateContext(ctxt->vadpy, ctxt->hwaccel->config_id, avctxt->width, avctxt->height, VA_PROGRESSIVE, ctxt->surface_ids, ctxt->surface_count, &ctxt->hwaccel->context_id)) != VA_STATUS_SUCCESS)
				printf("ERROR init_vo: create context\n");
		}		

		ret = ((ret >= 0) && (status == VA_STATUS_SUCCESS)) ? 1 : -1;

		ctxt->v_ratio = ratio;
		ctxt->v_width = avctxt->width;
		ctxt->v_height = avctxt->height;
		ctxt->csp = fmt;
		ctxt->imgfmt = imgfmt;
		ctxt->fmt_changed = 1;
	}

	if(ifmts)
		free(ifmts);

	if(ret < 0)
		free_vo(ctxt);

	printf("init_vo returning %d\n", ret);
	return ret;
}

static int query_format(ctxt_t *ctxt, int fmt, VAEntrypoint ep, VAProfile prfl)
{
	free_vo(ctxt);
	return (init_vo(ctxt, fmt, ep, prfl) >= 0) ? 1 : 0;
}

static enum PixelFormat get_format(struct AVCodecContext *avctxt, const enum PixelFormat *fmt)
{
	enum PixelFormat *cur;
	enum PixelFormat pfmt = PIX_FMT_NONE;
	VAEntrypoint ep = 0;
	VAProfile prfl = -1;
	int hw, ifmt;
	ctxt_t* ctxt = avctxt->opaque;

	ctxt->pixfmt_default = avcodec_default_get_format(avctxt, fmt);

	for(hw = 1; hw >= 0; hw--)
	{
		for(cur = (enum PixelFormat*)fmt; *cur != PIX_FMT_NONE; cur++)
		{
			ifmt = pixfmt2imgfmt(*cur, avctxt->codec_id);
			printf("trying pixfmt 0x%08X imgfmt 0x%08X  avctxt-pixfmt 0x%08X\n", *cur, ifmt, avctxt->pix_fmt);
			if((hw ^ ((ifmt & 0xFFFF0000) == IMGFMT_VAAPI)) != 0)
                continue;

			if(hw && (get_entrypoint(ctxt, &ep, &prfl) == 0))
				continue;

			if(query_format(ctxt, *cur, ep, prfl))
				break;
		}

		if(*cur != PIX_FMT_NONE)
			break;
	}	

	pfmt = *cur;
	if(pfmt == PIX_FMT_NONE)
	{
		printf("pixelformat get default...\n");
		pfmt = ctxt->pixfmt_default;
		if(query_format(ctxt, pfmt, ep, prfl) < 0)
			pfmt = PIX_FMT_NONE;
	}

	set_format_params(avctxt, pfmt);

	printf("pixelformat %d\n", pfmt);
	return pfmt;
}

#define INBUFFSIZE 1000000
#define BUFFSIZE 32768

#define PD_BUFF 4090
#define PD_BUFFMAX 4 * PD_BUFF



static int read_data(ctxt_t* this, int size)
{
	FILE* f = this->file;
	int ret = -1;
	
	unsigned char* data = malloc(size);

	int sz = fread(data, 1, size, f);
	int ds = sz;
	int sleft = (this->buf.inwpos - INBUFFSIZE) + sz;
	if(sleft < 0)
		sleft = 0;
	//printf("sz %d left %d wp %d ",sz,sleft,this->vd.inwpos);
	sz -= sleft;
	if(sz)
	{
		memcpy(&this->buf.indata[this->buf.inwpos], data, sz);
		this->buf.inwpos += sz;
		if(this->buf.inwpos == INBUFFSIZE)
			this->buf.inwpos = 0;
		if(sleft)
			memcpy(&this->buf.indata[this->buf.inwpos], &data[sz], sleft);
		this->buf.inwpos += sleft;
		this->buf.bsize += ds;
		ret = this->buf.bsize;
	}
	free(data);
	return ret;
}

static int64_t seek_buff(void *opaque, int64_t offset, int whence)
{
	dbuf_t* db = (dbuf_t*)opaque;
	ctxt_t* this = (ctxt_t*)db->this;

	FILE* f = this->file;
	int64_t ret = -1;
	
	if(!f)
		return ret;

	if(!this->seekable)
	{
		printf("seek buff: not seekable (offs %ld whence %d)\n", offset, whence);
		this->buf.inrpos = 0;
		return -1;
	}

	switch(whence)
	{
	case AVSEEK_SIZE:
		{
			int64_t p0 = ftell(f);
			fseek(f, 0, SEEK_END);
			ret = ftell(f);
			fseek(f, p0, SEEK_SET);
			break;
		}
	default: fseek(f, offset, whence); ret = ftell(f); reset_buf(this); break;
	}

	printf("seek: %ld (%ld)  %ld %d\n", ret, ftell(f), offset, whence);
	return ret;
}

#define LOG_READ
static int read_buff(void *opaque, uint8_t *buf, int buf_size)
{
	dbuf_t* db = (dbuf_t*)opaque;
	ctxt_t* this = (ctxt_t*)db->this;
	int sz = buf_size;

	if(!this)
		return 0;

	while(/*this->run && */(db->bsize < buf_size))
	{
		int r = read_data(this, buf_size);
		if((r >= 0) || (r == -EAGAIN))
		{
			if(db->bsize)
				break;
			//printf("no data, waiting...\n");
			usleep(10000);
		}
		else
		{
			printf("ERROR: read_data %d\n", r);
			//if(r == -1)
				//return -1;
			return r;
		}
	}

#if 1
	if(db->bsize < buf_size)
		buf_size = db->bsize;
#endif	

	//printf("read %p %d (%d)...\n", buf, buf_size, sz);

#ifdef LOG_READ
	printf("rd: sz %d ",sz);
#endif
	sz = buf_size;
	int sleft = (db->inrpos - INBUFFSIZE) + sz;
	if(sleft < 0)
		sleft = 0;
	sz -= sleft;
#ifdef LOG_READ
	printf("left %d rp %d ", sleft, db->inrpos);
#endif
	memcpy(buf, &db->indata[db->inrpos], sz);
	db->inrpos += sz;
	if(db->inrpos == INBUFFSIZE)
		db->inrpos = 0;
	if(sleft)
		memcpy(&buf[sz], &db->indata[db->inrpos], sleft);
	db->inrpos += sleft;
	db->bsize -= buf_size;
	
	db->pos += buf_size;
#ifdef LOG_READ
	printf("--> sz %d rp %d (-> %d) bsize %d\n", buf_size, db->inrpos, db->pos, db->bsize);
#endif

	if(!buf_size)
		return -EAGAIN;

	return buf_size;
}

static VAStatus get_avctxt(ctxt_t *ctxt, const char* fname)
{
	AVCodec* avcodec = NULL;
	VAStatus status = VA_STATUS_ERROR_UNKNOWN;
	int i;


#if 0
	if(av_open_input_file(&ctxt->ic, fname, NULL, 0, NULL) != 0)
	{
		printf("ERROR open file '%s'\n", fname);
		return VA_STATUS_ERROR_UNKNOWN;
	}
#else

	if(ctxt->file)
		fclose(ctxt->file);
	ctxt->file = fopen(fname, "r");
	if(!ctxt->file)
		return VA_STATUS_ERROR_UNKNOWN;

	AVProbeData pd;

	pd.filename = "";
	pd.buf_size = 0;
	pd.buf = NULL;
	if(!ctxt->format)
		pd.buf = malloc(PD_BUFFMAX);

	int sz = 0;

	if(!ctxt->format && pd.buf)
	{
		int rs = 0;

		for(;;)
		{
			rs = PD_BUFF;
			if((sz + rs) > ctxt->buf.bsize)
				read_data(ctxt, (sz + rs) - ctxt->buf.bsize);

			if((sz + rs > PD_BUFFMAX) || ((sz + rs) > ctxt->buf.bsize))
				break;
			memcpy(pd.buf, &ctxt->buf.indata[sz], rs);
			sz += rs;

			pd.buf_size = sz;
			//pd.buf_size = PD_BUFF;
			printf("find format sz %d buf %d\n", sz, ctxt->buf.bsize);
			ctxt->format = av_probe_input_format(&pd, 1);
			if(ctxt->format)
			{
				printf("demux: got format %s (%s)\n", ctxt->format->name, ctxt->format->long_name);			
				break;
			}
		}			
	}

	if(pd.buf)
		free(pd.buf);

	if(!ctxt->format)
	{
		printf("ERROR find format\n");
		return VA_STATUS_ERROR_UNKNOWN;
	}

	//seek_buff(&ctxt->buf, 0, SEEK_SET);
	//reset_buf(ctxt);

	if(!ctxt->ioctxt)
	{
		ctxt->ioctxt = av_alloc_put_byte(ctxt->buf.data, BUFFSIZE, 0, &ctxt->buf, read_buff, NULL, seek_buff);

		if(!ctxt->ioctxt)
		{
			printf("ERROR alloc_put_byte\n");
			return VA_STATUS_ERROR_UNKNOWN;
		}

		ctxt->ioctxt->is_streamed = !ctxt->seekable;
		if(!ctxt->ic)
		{
			printf("open input (streamed %d)...\n", ctxt->ioctxt->is_streamed);	
			if(av_open_input_stream(&ctxt->ic, ctxt->ioctxt, "", ctxt->format, NULL) < 0)
			{
				printf("ERROR av_open_input_stream\n");
				return VA_STATUS_ERROR_UNKNOWN;
			}
		}
	}

	ctxt->ioctxt->is_streamed = !ctxt->seekable;

#endif

	ctxt->ic->probesize = 262144;
	if(av_find_stream_info(ctxt->ic) < 0)
		printf("ERROR av_find_stream_info\n");

	dump_format(ctxt->ic, 0, "", 0);

	AVStream* vs = NULL;
	for(i = 0; i < ctxt->ic->nb_streams; i++)
	{
		if((ctxt->ic->streams[i]->codec->codec_type == CODEC_TYPE_VIDEO) && !vs)
			vs = ctxt->ic->streams[i];
		else
			ctxt->ic->streams[i]->discard = AVDISCARD_ALL;
	}
	
	printf("Stream #%d: w %d h %d\n", vs->index, vs->codec->width, vs->codec->height);

	ctxt->avctxt = vs->codec;
	if(!ctxt->avctxt)
		printf("ERROR get AVContext!\n");
		
	if(ctxt->avctxt)
	{
		ctxt->avctxt->opaque = ctxt;
		avcodec = avcodec_find_decoder(ctxt->ic->streams[vs->index]->codec->codec_id);
		if(!avcodec)
			printf("ERROR get AVCodec!\n");
	}

	char dr1 = 0;
	if(ctxt->avctxt && avcodec)
	{
		if((avcodec->capabilities & CODEC_CAP_DR1) && (avcodec->id != CODEC_ID_H264) && (avcodec->id != CODEC_ID_INTERPLAY_VIDEO) && (avcodec->id != CODEC_ID_ROQ))
			dr1 = 1;

		ctxt->avctxt->get_format = get_format;
		
		if(dr1)
			ctxt->avctxt->flags |= CODEC_FLAG_EMU_EDGE;
		
		status = VA_STATUS_SUCCESS;
	}

	if(avcodec)
	{
		printf("AvCodecContext %p w %d h %d codec: %d '%s' dr1: %d\n", ctxt->avctxt, ctxt->avctxt->width, ctxt->avctxt->height, avcodec->id, avcodec->long_name ? avcodec->long_name : avcodec->name, dr1);


		if(avcodec_open(ctxt->avctxt, avcodec) < 0)
		{
			printf("ERROR open AVCodec!\n");
			status = VA_STATUS_ERROR_UNKNOWN;
		}

		ctxt->avctxt->dsp_mask = FF_MM_FORCE | FF_MM_MMX | FF_MM_MMXEXT | FF_MM_SSE;

		if(!ctxt->hwaccel)
			ctxt->hwaccel = (struct vaapi_context*)calloc(1, sizeof(struct vaapi_context));
		ctxt->hwaccel->display = ctxt->vadpy;
		ctxt->avctxt->hwaccel_context = ctxt->hwaccel;

		if(!ctxt->avframe)
			ctxt->avframe = avcodec_alloc_frame();
		if(!ctxt->avframe)
		{
			status = VA_STATUS_ERROR_UNKNOWN;
			printf("ERROR allocate avframe!\n");
		}

		
	}
//seek_buff(&ctxt->buf, 0, SEEK_SET);
	
	return status;
}


int init_wnd(Display** dpy, VADisplay **vadpy, Window* wnd, int *w, int *h);
void destroy_wnd(Display** dpy, VADisplay **vadpy, Window* wnd);

void playfile(const char* fname, Display** dpy, VADisplay** vadpy, Drawable* drawable, int wnd_w, int wnd_h)
{
	ctxt_t ctxt;

	VAStatus status = VA_STATUS_ERROR_UNKNOWN;

	ctxt.fatal_error = 0;
	ctxt.vadpy = *vadpy;
	ctxt.wnd_w = wnd_w;
	ctxt.wnd_h = wnd_h;
	ctxt.hwaccel = NULL;
	ctxt.avctxt = NULL;
	ctxt.avframe = NULL;
	ctxt.surface_count = 0;
	ctxt.surfaces = NULL;
	ctxt.surface_ids = NULL;
	ctxt.surface_idx = 0;
	ctxt.sw_surface = NULL;
	ctxt.v_width = 0;
	ctxt.v_height = 0;
	ctxt.v_ratio = 0.0;
	ctxt.csp = -1;
	ctxt.imgfmt = -1;
	ctxt.pixfmt_default = PIX_FMT_NONE;
	ctxt.va_imgfmt.fourcc = 0;
	ctxt.fmt_changed = 0;
	ctxt.hw = 0;
	ctxt.ic = NULL;
	ctxt.ioctxt = NULL;
	ctxt.format = NULL;
	ctxt.file = NULL;
	ctxt.buf.this = &ctxt;
	ctxt.buf.data = malloc(BUFFSIZE);
	ctxt.buf.indata = malloc(INBUFFSIZE);
	ctxt.seekable = 1;

	reset_buf(&ctxt);

	av_register_all();
	
	AVPacket avpkt;
	int pic, len, i, deint;
	XRectangle rsrc = { 0, 0 }, rdst = { 0, 0 };
	
	deint = 2;

	av_init_packet(&avpkt);

	struct termios term, oterm;
	int fd = 0;
	tcgetattr(fd, &oterm);
	memcpy(&term, &oterm, sizeof(term));
	term.c_lflag = term.c_lflag & ~(ICANON | ECHO);
	term.c_cc[VMIN] = 0;
	term.c_cc[VTIME] = 0;
	tcsetattr(fd, TCSANOW, &term);
	
	int clxs = 0, clys = 0, clxd = 0, clyd = 0;
	
	int zoom = 100;
	int z = -1;
	int pause = 0;
	int printerl = 0;

	struct timeval t1, t2;
	int diff = 0;

	status = get_avctxt(&ctxt, fname);
	if(status == VA_STATUS_SUCCESS)
	{
		printf("starting loop...\n");
		while(!ctxt.fatal_error && !quit_request && (pause || av_read_frame(ctxt.ic, &avpkt)) >= 0)
		{
			gettimeofday(&t1, NULL);
			//av_log_set_level(48);

//#define TEST1
#ifdef TEST1
			static int y = 0;
			if(ctxt.surface_count && !pause)
			{
				int scnt = ctxt.surface_count;
				if(y == 10)
				{
					int i;
					for(i = 0; i < ctxt.surface_count; i++)
					{
						printf("+++++++++++++ destr. surfaces ++++++++++   0x%08X 0x%08X\n", ctxt.surfaces[i].id, ctxt.surfaces[i].img ? ctxt.surfaces[i].img->image_id : -1);
						if(ctxt.surfaces[i].img)
							printf("vaDestroyImage %d\n", vaDestroyImage(ctxt.vadpy, ctxt.surfaces[i].img->image_id));
						printf("vaDestroySurfaces %d\n", vaDestroySurfaces(ctxt.vadpy, &ctxt.surfaces[i].id, 1));
						ctxt.surface_count = 0;
						if(ctxt.surfaces[i].img)
							free(ctxt.surfaces[i].img);
						ctxt.surfaces[i].img = NULL;
						ctxt.surfaces[i].id = -1;
					}
					y = 0;
				}

				if(ctxt.surface_count == 0)
				{
					ctxt.surface_count = scnt;
					int i;
					for(i = 0; i < ctxt.surface_count; i++)
					{
						ctxt.surfaces[i].used = 0;
						ctxt.surfaces[i].bound = 0;
						if(!ctxt.hw)
							ctxt.surfaces[i].img = malloc(sizeof(VAImage));
						ctxt.surfaces[i].id = ctxt.surfaces[1].id;
						printf("---> create srfc: %d\n", vaCreateSurfaces(ctxt.vadpy, ctxt.avctxt->width, ctxt.avctxt->height, VA_RT_FORMAT_YUV420, 1, &ctxt.surfaces[i].id));
						if(!ctxt.hw)
						{
							printf("---> create img dpy %p %dx%d: %d\n", ctxt.vadpy, ctxt.avctxt->width, ctxt.avctxt->height, vaCreateImage(ctxt.vadpy, &ctxt.va_imgfmt, ctxt.avctxt->width, ctxt.avctxt->height, ctxt.surfaces[i].img));
							printf("---> 0x%08X 0x%08X %dx%d\n", ctxt.surfaces[i].id, ctxt.surfaces[0].img->image_id, ctxt.surfaces[i].img->width, ctxt.surfaces[i].img->height);
						}
					}
					ctxt.sw_surface = &ctxt.surfaces[0];
				}
				y++;
			}

#endif

			pic = 0;
			len = -1;
			if(!pause && (avpkt.stream_index == 0))
				len = avcodec_decode_video2(ctxt.avctxt, ctxt.avframe, &pic, &avpkt);


#if 1
			printf("packet readed decode: %5d pause %d ispic %2d cnr %5d stream %d p.size %5d surf 0x%08X psurf %8p pts %ld ratio %d %d dur %d\n",
				len, pause, pic, ctxt.avframe ? ctxt.avframe->coded_picture_number : -1, avpkt.stream_index, avpkt.size, ctxt.avframe ? (int)(size_t)ctxt.avframe->data[3] : -1, ctxt.avframe ? ctxt.avframe->data[0] : NULL, avpkt.pts, ctxt.avctxt->sample_aspect_ratio.num, ctxt.avctxt->sample_aspect_ratio.den, avpkt.duration);
#endif


			if((z != zoom) || ctxt.fmt_changed)
			{
				//printf("zoom %d\n", zoom);

				clyd = 0;
				clxd = 0;

				int h = wnd_h * zoom / 100.0;

				clxs = -(wnd_w - (h * ctxt.v_ratio)) / 2;
				clys = -(wnd_h - (wnd_w - 2 * -clxs) / ctxt.v_ratio) / 2;

				printf("cls   wnd %d x %d  r %f  %d %d cld %d %d  %d %d %d %d -> %d %d %d %d\n", wnd_w, wnd_h, ctxt.v_ratio, clxs, clys, clxd, clyd, rsrc.x, rsrc.y, rsrc.width, rsrc.height, rdst.x, rdst.y, rdst.width, rdst.height);

				if(clxs < 0)
				{
					clxd = -clxs;
					clxs = 0;
				}
				
				if(clys < 0)
				{
					clyd = -clys;
					clys = 0;
				}

				rsrc.x = clxs;
				rsrc.y = clys;
				rsrc.width = ctxt.avctxt->width - 2 * clxs;
				rsrc.height = ctxt.avctxt->height - 2 * clys;

				rdst.x = clxd;
				rdst.y = clyd;
				rdst.width = wnd_w - 2 * clxd;
				rdst.height = wnd_h - 2 * clyd;
			}

			z = zoom;
			ctxt.fmt_changed = 0;
		
			if(pic)
			{
				if(!ctxt.hw)
					draw_slice(&ctxt, ctxt.avframe->data, ctxt.avframe->linesize, ctxt.avctxt->width, ctxt.avctxt->height, 0, 0);

				surface_t* srfc = NULL;
				if(ctxt.hw)
				{
					srfc = (surface_t*)ctxt.avframe->data[0];
				}
				else
				{
					srfc = ctxt.sw_surface;
					if(!srfc->bound)
					{
						int w = ctxt.avctxt->width;
						int h = ctxt.avctxt->height;
						if((status = vaPutImage(ctxt.vadpy, srfc->id, srfc->img->image_id, 0, 0, w, h, 0, 0, w, h)) != VA_STATUS_SUCCESS)
						{
							printf("ERROR vaPutImage: %d\n", status);
						}
					}
					srfc->used = 0;
				}

				if(printerl)
					printf("interlaced: %d dur. %d\n", ctxt.avframe->interlaced_frame, avpkt.duration);


				//deint=0; //xvba!

				for(i = 0; i <= !!((deint > 1) && ctxt.avframe->interlaced_frame); i++)
				{
					unsigned int flags = (deint && (ctxt.avframe->interlaced_frame) ? (((!!(ctxt.avframe->top_field_first)) ^ i) == 0 ? VA_BOTTOM_FIELD : VA_TOP_FIELD) : VA_FRAME_PICTURE);
//printf("Putsrfc srfc 0x%08X  wnd %d, 5%d, %d %dx%d -> %d, %d, %dx%d\n", srfc->id, (int)*drawable, rsrc.x, rsrc.y, rsrc.width, rsrc.height, rdst.x, rdst.y, rdst.width, rdst.height);
					if((status = vaPutSurface(ctxt.vadpy, srfc->id, *drawable, rsrc.x, rsrc.y, rsrc.width, rsrc.height,
						rdst.x, rdst.y, rdst.width, rdst.height, NULL, 0, flags)) != VA_STATUS_SUCCESS)
					{
						printf("ERROR vaPutSurface (vadpy 0x%08X, surface 0x%08X, window %d, flags 0x%08X): %d\n", (int)(size_t)*ctxt.vadpy, srfc->id, (int)*drawable, flags, status);
						break;
					}
				}

				gettimeofday(&t2, NULL);
				diff = (t2.tv_sec > t1.tv_sec) ? t2.tv_usec + (1000000 - t1.tv_usec) : t2.tv_usec - t1.tv_usec;

				int dur = 0;
				switch(avpkt.duration)
				{
				case 0: dur = ctxt.avctxt->time_base.num / (ctxt.avctxt->time_base.den / 1000000) * 2; break;
				case 1: dur = (ctxt.avctxt->time_base.num * 1000000) / ctxt.avctxt->time_base.den * 2; break;
				default: dur = avpkt.duration * 100 / 9;
				}

				int ts = dur - diff;
				printf("t decode %d base %d/%d dur %d -> %d ts %d\n",diff, ctxt.avctxt->time_base.num, ctxt.avctxt->time_base.den, avpkt.duration, dur, ts);
				if(ts > 0)
					usleep(ts);
			}

			av_free_packet(&avpkt);

			int k = getchar();

//#define TEST2
#ifdef TEST2
			static int z=0;
			if(z == 30)
			{
				k = 'r';
				z = 0;
			}
			z++;
#endif

			switch(k)
			{
			case 'Q':
			case 'q': quit_request = 1; break;
			case ' ': pause = !pause; break;
			case '1': rsrc.x += OFFSDELTA; rsrc.width -= OFFSDELTA * 2; break;
			case '2': rsrc.x -= OFFSDELTA; rsrc.width += OFFSDELTA * 2; break;
			case '3': rsrc.x += OFFSDELTA; break;
			case '4': rsrc.x -= OFFSDELTA; break;
			case '5': rsrc.width += OFFSDELTA * 2; break;
			case '6': rsrc.width -= OFFSDELTA * 2; break;
			case 'i': printerl = !printerl; break;
			case 'z': zoom += 5; break;
			case 'Z': zoom -= 5; break;
			case 'd':
			case 'D': deint++; if(deint > 2) deint = 0; printf("deinterlacer now: %d\n", deint); break;
			case 'r':
				close_avctxt(&ctxt);
				destroy_surfaces(&ctxt);
				free_vo(&ctxt);

#if 1
#if 1
				destroy_wnd(dpy, vadpy, drawable);
				init_wnd(dpy, vadpy, drawable, &wnd_w, &wnd_h);
#else
				vaTerminate(vadpy);
				int maj, min;
				vaInitialize(vadpy, &maj, &min);
#endif
#endif
				ctxt.vadpy = *vadpy;
				
				get_avctxt(&ctxt, fname);
				break;
			default:;
			}

			if(len == 0)
				break;

		}

	}
	else
		printf("get_avctxt: %d\n", status);

	tcsetattr(fd, TCSANOW, &oterm);

	av_free_packet(&avpkt);
	

	close_avctxt(&ctxt);
	destroy_surfaces(&ctxt);
	free_vo(&ctxt);

	if(ctxt.hwaccel)
		free(ctxt.hwaccel);
	ctxt.hwaccel = NULL;

	if(ctxt.buf.data)
		free(ctxt.buf.data);
	if(ctxt.buf.indata)
		free(ctxt.buf.indata);

		
	printf("playfile done.\n");

}

int x11_error_handler(Display *dpy, XErrorEvent *error)
{
	printf("X11-Error %d\n", error->error_code);
	x11_error = True;
	return 0;
}

void x11_wait_event(Display *dpy, Window w, int event)
{
	XEvent e;
	do
	{
		XMaskEvent(dpy, (event == Expose) ? ExposureMask : StructureNotifyMask, &e);
	}
	while((e.type != event) || ((event == Expose) ? (e.xexpose.window != w) : (e.xmap.event != w)));
}

static const int x11_event_mask = 
	(KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask | EnterWindowMask |
	ExposureMask | StructureNotifyMask);
	

#define VAINITONCE

int init_wnd(Display** dpy, VADisplay **vadpy, Window* wnd, int *w, int *h)
{
	//*dpy = XOpenDisplay(NULL);
	printf("XDisplay: %p\n", *dpy);
	
	if(*dpy)
	{
		XWindowAttributes wattr;
		XSetWindowAttributes sattr;
		unsigned long sattr_mask;
		int err = 0;
		
		int screen = DefaultScreen(*dpy);
		
		int root_window = RootWindow(*dpy, screen);
		unsigned long black_pixel = BlackPixel(*dpy, screen);
		unsigned long white_pixel = WhitePixel(*dpy, screen);
		
		XGetWindowAttributes(*dpy, root_window, &wattr);
		int depth = wattr.depth;
		if((depth != 15) && (depth != 16) && (depth != 24) && (depth != 32))
			depth = 24;
		
		XVisualInfo visualInfo;
		XVisualInfo* vi = &visualInfo;
		XMatchVisualInfo(*dpy, screen, depth, TrueColor, vi);
		Colormap cmap = CopyFromParent;
		
		sattr_mask = CWBorderPixel | CWBackPixel;
		sattr.border_pixel = black_pixel;
		sattr.background_pixel = white_pixel;
		sattr.override_redirect = 1;

		if(cmap != CopyFromParent)
		{
			sattr_mask |= CWColormap;
			sattr.colormap = cmap;
		}
		
		*w = wattr.width;
		*h = wattr.height;
		
		printf("creating window w %d h %d\n",*w, *h);
		*wnd = XCreateWindow(*dpy, root_window, 0, 0, *w, *h, 0, depth, InputOutput, vi->visual, sattr_mask, &sattr);
		if(wnd == None)
		{
			printf("ERROR create window\n");
			err = 1;
		}

		if(!err)
		{
			printf("window: %d\n", (int)*wnd);
		
			XSelectInput(*dpy, *wnd, x11_event_mask);
		
			XMapWindow(*dpy, *wnd);
			x11_wait_event(*dpy, *wnd, MapNotify);
			x11_wait_event(*dpy, *wnd, Expose);
		}	
		if(!err)
		{
#ifndef VAINITONCE
			*vadpy = vaGetDisplay(*dpy);
			if(*vadpy)
			{
				int maj, min;
				vaInitialize(*vadpy, &maj, &min);
				printf("libva: dpy 0x%08X   %d.%d\n", (int)(size_t)*vadpy, maj, min);
				return 1;
			}
#else
			return 1;
#endif
		}
	}
	return 0;
}

void destroy_wnd(Display** dpy, VADisplay **vadpy, Window* wnd)
{
#ifndef VAINITONCE
	printf("terminate vadpy 0x%08X\n", (int)(size_t)*vadpy);
	if(*vadpy)
		vaTerminate(*vadpy);
	*vadpy = NULL;
#endif
	if(*wnd)
	{	
		XUnmapWindow(*dpy, *wnd);
		x11_wait_event(*dpy, *wnd, UnmapNotify);
		XDestroyWindow(*dpy, *wnd);
	}

	//XCloseDisplay(*dpy);
	
	//*dpy = NULL;
	*wnd = 0;
}

int main(int argc, char* argv[])
{
	Display* dpy = NULL;
	VADisplay* vadpy = NULL;
	Window wnd = 0;
	const char* fname = NULL;
	XErrorHandler x11_old_err_handler = NULL;

	x11_old_err_handler = XSetErrorHandler(x11_error_handler);
	x11_error = False;

	signal(SIGTERM,exit_sighandler);
	signal(SIGHUP,exit_sighandler);
	signal(SIGINT,exit_sighandler);
	signal(SIGQUIT,exit_sighandler);

	if(argc < 2)
		printf("\nno file given!\n\n");
	else
		fname = argv[1];
	if(fname)
		printf("\nwill playing file: '%s'\n\n", fname);

dpy = XOpenDisplay(NULL);

	int w, h;
	if(init_wnd(&dpy, &vadpy, &wnd, &w, &h))
	{
#ifdef VAINITONCE
		vadpy = vaGetDisplay(dpy);
		if(vadpy)
		{
			int maj, min;
			vaInitialize(vadpy, &maj, &min);
			printf("libva: dpy 0x%08X   %d.%d\n", (int)(size_t)vadpy, maj, min);
		}
#endif
		playfile(fname, &dpy, &vadpy, &wnd, w, h);
	}

#ifdef VAINITONCE
	if(vadpy)
		vaTerminate(vadpy);
#endif
	destroy_wnd(&dpy, &vadpy, &wnd);

XCloseDisplay(dpy);

	XSetErrorHandler(x11_old_err_handler);

	return 0;
}

